unit DirectSoundOut;

{ DirectSound wav output (c)2003 by Paul TOTH <tothpaul@free.fr> }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

{-$define log}
{-$define setformat}
interface

uses
 Forms,Windows,MMSystem,SysUtils,OBuffer,Header;

const
  TWO_TIMES  = 5;
  BUFFERSIZE = OBUFFERSIZE shl TWO_TIMES;
  BIT_SELECT = $1f;
  SLEEPTIME  = 20;

const
  DSSCL_NORMAL                =1;
  DSSCL_PRIORITY              =2;
  DSBCAPS_PRIMARYBUFFER       =$00000001;
  DSBCAPS_STICKYFOCUS         =$00004000;
  DSBCAPS_GETCURRENTPOSITION2 =$00010000;  // More accurate play cursor under emulation
  DSBPLAY_LOOPING             =$00000001;
  DSBLOCK_FROMWRITECURSOR     =$00000001;

type
 DSBUFFERDESC = record
  dwSize: cardinal;
  dwFlags: cardinal ;
  dwBufferBytes: cardinal;
  dwReserved: cardinal;
  lpwfxFormat: PWaveFormatEx;
 end;

 IDirectSoundBuffer=Interface
  procedure _GetCaps;
  function GetCurrentPosition(var dwCurrentPlayCursor,lpdwCurrentWriteCursor:cardinal):HResult; stdcall;
  procedure _GetFormat;
  procedure _GetVolume;
  procedure _GetPan;
  procedure _GetFrequency;
  procedure _GetStatus;
  procedure _Initialize;
  function Lock(Offset,Len:cardinal; var p1:pointer; var l1:cardinal; var p2:pointer; var l2:cardinal; Flags:cardinal):HResult; stdcall;
  function Play(dwReserved1, dwReserved2, dwFlags:cardinal):HResult; stdcall;
  procedure _SetCurrentPosition;
  function SetFormat(const Format:TWaveFormatEx):HResult; stdcall;
  procedure _SetVolume;
  procedure _SetPan;
  procedure _SetFrequency;
  function Stop:HResult; stdcall;
  function UnLock(p1:pointer; l1:cardinal; p2:pointer; l2:cardinal):HResult; stdcall;
  function Restore:HResult; stdcall;
 end;


 IDirectSound=Interface
  function CreateSoundBuffer(const lpDSBufferDesc: DSBUFFERDESC; var lplpDirectSoundBuffer:IDirectSoundBuffer; pUnkOuter:IUnknown ): HRESULT; stdcall;
  procedure _GetCaps;
  procedure _DuplicateSoundBuffer;
  function SetCooperativeLevel(hWnd:HWND; dwFlags:DWORD):HRESULT; stdcall;
 end;

 TDirectSoundOut=class(TOBuffer)
 private
  fSoundBuffer :IDirectSoundBuffer;
  fChannels    :integer;
  fFrequency   :integer;   // define Buffer size
  fDataSize    :cardinal;  // Buffer size
  fData        :PChar;
  fOutputs     :array {channels} of PWord;
  fFillup      :integer;
  fBufferCount :integer;
  fBufferSize  :cardinal;
  fBufferOffset:cardinal;
  procedure ClearSoundBuffer;
 public
  constructor Create;
  destructor Destroy; override;
  procedure SetFormat(Channels,Frequency:integer; Version:TVersion; Layer:integer); override;
  procedure NoMoreData; override;
  procedure Append(Channel: Cardinal; Value: SmallInt); override;
  procedure WriteBuffer; override;
  procedure ClearBuffer; override;
  procedure SetStopFlag; override;
 end;

implementation


function DirectSoundCreate(lpGuid:PGUID; var ppDS:IDirectSound; pUnkOuter:IUnknown ):HRESULT; stdcall; external 'DSOUND.DLL';

procedure DSCheck(Value:HResult; Method:string);
var
 s:string;
begin
 if Value=0 then exit;
 case cardinal(Value) of
  $80070057: s:='DSERR_INVALIDPARAM  - An invalid parameter was passed to the returning function.';
  else       s:='$'+IntToHex(Value,8)
 end;
 raise Exception.Create('['+Method+'] DirectSound Error '+s);
end;

var
 gDirectSound:IDirectSound=nil;
{$ifdef setformat}
 gSoundBuffer:IDirectSoundBuffer;
 gBufferCount:integer;
{$endif}

constructor TDirectSoundOut.Create;
{$ifdef setformat}
var
 BD:DSBUFFERDESC;
{$endif}
begin
 if gDirectSound=nil then begin
  DirectSoundCreate(nil,gDirectSound,nil);
  if gDirectSound=nil then raise Exception.Create('DirectSound general error');
  DSCheck(gDirectSound.SetCooperativeLevel(Application.Handle, DSSCL_PRIORITY),'SetCooperativeLevel');
 {$ifdef setformat}
  FillChar(BD,SizeOf(BD),0);
  BD.dwSize :=SizeOf(BD);
  BD.dwFlags:=DSBCAPS_PRIMARYBUFFER;
  DSCheck(gDirectSound.CreateSoundBuffer(BD,gSoundBuffer,nil),'CreateSoundBuffer (primary)');
  gBufferCount:=0;
 {$endif}
 end;
end;

destructor TDirectSoundOut.Destroy;
begin
 if fSoundBuffer<>nil then begin
  fSoundBuffer.Stop;
  fSoundBuffer:=nil;
 end;
 inherited;
end;

procedure TDirectSoundOut.SetFormat(Channels,Frequency:integer; Version:TVersion; Layer:integer);
var
 WF:TWaveFormatEx;
 BD:DSBUFFERDESC;
 DataSize:cardinal;
 i:integer;
begin
 DataSize:=Channels*BUFFERSIZE;
 if (Version=MPEG2_LSF) then DataSize:=DataSize shr 1;
 if (Layer  =1        ) then DataSize:=DataSize div 3;
 if DataSize<>fDataSize then begin
  if fDataSize<>0 then FreeMem(fData);
  fDataSize:=DataSize;
  GetMem(fData,fDataSize);
  fBufferSize:=3*fDataSize;
 end;
 if (fSoundBuffer=nil)or(Channels<>fChannels)or(Frequency<>fFrequency) then begin
  fChannels :=Channels;
  fFrequency:=Frequency;
  WF.wBitsPerSample  := 16;  // No 8-bit support yet
  WF.wFormatTag      := WAVE_FORMAT_PCM;
  WF.nChannels       := fChannels;
  WF.nSamplesPerSec  := fFrequency;
  WF.nAvgBytesPerSec := fChannels * fFrequency shl 1;
  WF.nBlockAlign     := (fChannels shl 1);
  WF.cbSize          := 0;
 {$ifdef setformat}
  if fSoundBuffer<>nil then begin
   fSoundBuffer:=nil;
   dec(gBufferCount);
  end;
  if gBufferCount=0 then DSCheck(gSoundBuffer.SetFormat(WF),'SetFormat (primary)');
  inc(gBufferCount);
 {$endif}
  FillChar(BD,SizeOf(BD),0);
  BD.dwSize :=SizeOf(BD);
  BD.dwFlags:=DSBCAPS_GETCURRENTPOSITION2 or DSBCAPS_STICKYFOCUS;
  BD.dwBufferBytes:=fBufferSize;
  BD.lpwfxFormat:=@WF;
  DSCheck(gDirectSound.CreateSoundBuffer(BD,fSoundBuffer,nil),'CreateSoundBuffer');
 end;
 SetLength(fOutputs,fChannels);
 for i:=0 to fChannels-1 do begin
  fOutputs[i]:=@fData[i*fChannels];
 end;
 fFillup:=0;
 fBufferCount:=0;
 fBufferOffset:=0;
// DSCheck(fSoundBuffer.Play(0, 0, DSBPLAY_LOOPING),'Play');
end;

procedure TDirectSoundOut.Append(Channel: Cardinal; Value: SmallInt);
begin
//{$ifdef log}writeln(Channel);{$endif}
 fOutputs[Channel]^:=Value;
 inc(fOutputs[Channel],fChannels);
end;

procedure TDirectSoundOut.ClearSoundBuffer;
var
 PlayPos,WritePos:cardinal;
 p1,p2:pointer;
 l1,l2:cardinal;
begin
 DSCheck(fSoundBuffer.GetCurrentPosition(PlayPos,WritePos),'GetCurrentPosition');
 while (PlayPos>=fBufferOffset) and (PlayPos<fBufferOffset+fDataSize) do begin
  Sleep(10);
  DSCheck(fSoundBuffer.GetCurrentPosition(PlayPos,WritePos),'GetCurrentPosition');
 end;
 DSCheck(fSoundBuffer.Lock(fBufferOffset,fDataSize,p1,l1,p2,l2,0),'Lock');
 try
  if (p1<>nil)and(l1>0) then FillChar(p1^,l1,0);
  if (p2<>nil)and(l2>0) then FillChar(p2^,l2,0);
 finally
  DSCheck(fSoundBuffer.Unlock(p1,l1,p2,l2),'UnLock');
 end;
 inc(fBufferOffset,fDataSize);
 if fBufferOffset>=fBufferSize then fBufferOffset:=0;
end;

procedure TDirectSoundOut.NoMoreData;
begin
 if fFillUp>2 then begin
  if fSoundBuffer=nil then exit;
  ClearSoundBuffer;
  ClearSoundBuffer;
  ClearSoundBuffer;
  fFillup:=0;
 end;
end;

procedure TDirectSoundOut.WriteBuffer;
var
 PlayPos,WritePos:cardinal;
 p1,p2:pointer;
 l1,l2:cardinal;
 i:integer;
begin
// fSoundBuffer.Lock()
//{$ifdef log}WriteLn(fBufferCount);{$endif}
 inc(fBufferCount);
 if (fBufferCount and BIT_SELECT = 0) then begin
  fBufferCount := 0;

  for i:=0 to fChannels-1 do begin
   fOutputs[i]:=@fData[i*fChannels];
  end;

  if fFillUp>2 then begin
   DSCheck(fSoundBuffer.GetCurrentPosition(PlayPos,WritePos),'GetCurrentPosition');
   while (PlayPos>=fBufferOffset) and (PlayPos<fBufferOffset+fDataSize) do begin
    Sleep(10);
    DSCheck(fSoundBuffer.GetCurrentPosition(PlayPos,WritePos),'GetCurrentPosition');
   end;
  end;

  DSCheck(fSoundBuffer.Lock(fBufferOffset,fDataSize,p1,l1,p2,l2,0),'Lock');
  try
   if (p1<>nil)and(l1>0) then Move(fData[ 0],p1^,l1);
   if (p2<>nil)and(l2>0) then Move(fData[l1],p2^,l2);
  finally
   DSCheck(fSoundBuffer.Unlock(p1,l1,p2,l2),'UnLock');
  end;
  inc(fBufferOffset,fDataSize);
  if fBufferOffset>=fBufferSize then fBufferOffset:=0;

  if fFillUp=2 then DSCheck(fSoundBuffer.Play(0, 0, DSBPLAY_LOOPING),'Play');
  inc(fFillUp);
 end;
end;

procedure TDirectSoundOut.ClearBuffer;
var
 i:integer;
begin
 fSoundBuffer.Stop;
 FillChar(fData^,fDataSize,0);
 for i:=0 to fChannels-1 do begin
  fOutputs[i]:=@fData[i*fChannels];
 end;
end;

procedure TDirectSoundOut.SetStopFlag;
begin
 fSoundBuffer.Stop;
end;

initialization
{$ifdef log}AllocConsole;{$endif}
finalization
{$ifdef setformat}
 gSoundBuffer:=nil;
{$endif} 
 gDirectSound:=nil;
end.
