unit WaveOut;

{ WaveOut (c)2003 by Paul TOTH <tothpaul@free.fr> }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

{-$define log}
interface

uses
 Windows,SysUtils,MMSystem,Header,OBuffer,Player;

const
  TWO_TIMES  = 5;
  BUFFERSIZE = OBUFFERSIZE shl TWO_TIMES;
  BIT_SELECT = $1f;
  SLEEPTIME  = 20;

type
 TWaveOut=class(TOBuffer)
 private
  fHandle    :HWaveOut;  // MCI Handle
  fChannels  :integer;   // # of channels
  fOutputs   :array {Channels} of PWord; // current output pointers
  fFrequency :integer;   // define Buffer size
  fVersion   :TVersion;  // same
  fLayer     :integer;   // same
  fDataSize  :cardinal;  // Buffer size
  fHeaders   :array[0..2] of TWaveHdr; // 3 buffers
  fWave1     :PWaveHdr;  // first one
  fWave2     :PWaveHdr;  // second one
  fWaveHdr   :PWaveHdr;  // last one (current)
  fBuffer    :integer;   // active buffer index
  fFillUp    :integer;   // startup sequence (need 2 fillup)
  fBufferCount:integer;  // buffer filling
  fUserStop  :integer;
  procedure SwapBuffers; // ring the 3 buffers
 public
  destructor Destroy; override;
  procedure SetFormat(Channels,Frequency:integer; Version:TVersion; Layer:integer); override;
  procedure Append(Channel: Cardinal; Value: SmallInt); override;
  procedure WriteBuffer; override;
  procedure ClearBuffer; override;
  procedure SetStopFlag; override;
 end;

implementation

procedure waveOutProc(Handle:HWaveOut; Msg:integer; WaveOut:TWaveOut; WaveHdr:PWaveHdr; dwReserved:integer); stdcall;
begin
 case Msg of
  WOM_CLOSE: ;
  WOM_DONE : ;
  WOM_OPEN : ;
 end;
end;

destructor TWaveOut.Destroy;
var
 i:integer;
begin
 if fHandle<>0 then begin
  waveOutReset(fHandle);
  waveOutClose(fHandle);
 end;
 if fDataSize<>0 then
  for i:=0 to 2 do
   FreeMem(fHeaders[i].lpData);
 inherited;
end;

procedure TWaveOut.SetFormat(Channels,Frequency:integer; Version:TVersion; Layer:integer);
var
 WF:TWaveFormatEx;
 DataSize:cardinal;
 i:integer;
begin
{$ifdef log}WriteLn('SetFormat()');{$endif}
// if (Channels=fChannels)and(Frequency=fFrequency)and(Version=fVersion)and(Layer=fLayer) then exit;
 fVersion  :=Version;
 fLayer    :=Layer;

 if (fHandle=0)or(Channels<>fChannels)or(Frequency<>fFrequency) then begin
  fChannels :=Channels;
  fFrequency:=Frequency;

  if fHandle<>0 then begin
   waveOutReset(fHandle);
   waveOutClose(fHandle);
   fHandle:=0;
  end;

  WF.wBitsPerSample  := 16;  // No 8-bit support yet
  WF.wFormatTag      := WAVE_FORMAT_PCM;
  WF.nChannels       := fChannels;
  WF.nSamplesPerSec  := fFrequency;
  WF.nAvgBytesPerSec := fChannels * fFrequency shl 1;
  WF.nBlockAlign     := (fChannels shl 1);
  WF.cbSize          := 0;

 // Find a waveOut device and open it
  for i:=0 to waveOutGetNumDevs-1 do begin
   if waveOutOpen(@fHandle,WAVE_MAPPER,@WF,0,0,CALLBACK_NULL)=MMSYSERR_NOERROR then break;
  end;
  if fHandle=0 then
//  if waveOutOpen(@fHandle, WAVE_MAPPER, @WF, 0, 0, WAVE_ALLOWSYNC) <> MMSYSERR_NOERROR then
   raise Exception.Create('Output device failure');
 end;

 DataSize:=fChannels*BUFFERSIZE;
 if (fVersion=MPEG2_LSF) then DataSize:=DataSize shr 1;
 if (fLayer  =1        ) then DataSize:=DataSize div 3;

 if DataSize<>fDataSize then begin
  if fDataSize<>0 then
   for i:=0 to 2 do
    FreeMem(fHeaders[i].lpData);
  fDataSize:=DataSize;
  FillChar(fHeaders,SizeOf(fHeaders),0);
  for i:=0 to 2 do begin
   GetMem(fHeaders[i].lpData,fDataSize);
   fHeaders[i].dwBufferLength:=fDataSize;
  end;
 end;

 fBuffer:=0;
 fFillup:=0;
 fUserStop:=0;
 fBufferCount:=0;

 SetLength(fOutputs,fChannels);
 SwapBuffers;
 ClearBuffer;
end;

procedure TWaveOut.Append(Channel: Cardinal; Value: SmallInt);
begin
 fOutputs[Channel]^:=Value;
 inc(fOutputs[Channel],fChannels);
end;

procedure TWaveOut.SwapBuffers;
var
 i:integer;
begin
{$ifdef log}WriteLn('SwapBuffers');{$endif}
 inc(fBuffer);
 if fBuffer>2 then fBuffer:=0;
 fWave1  :=fWave2;
 fWave2  :=fWaveHdr;
 fWaveHdr:=@fHeaders[fBuffer];
 for i:=0 to fChannels-1 do begin
  fOutputs[i]:=@fWaveHdr.lpData[i*fChannels];
 end;
end;

procedure TWaveOut.WriteBuffer;
begin
{$ifdef log}WriteLn('WriteBuffer [',fBufferCount,']');{$endif}
  inc(fBufferCount);
  if (fBufferCount and BIT_SELECT = 0) then begin
   fBufferCount := 0;
   // Wait for 2 completed headers
   if (fFillup > 1) then begin
    // Prepare & write newest header
    waveOutPrepareHeader(fHandle,fWaveHdr,SizeOf(TWaveHdr));
    waveOutWrite(fHandle,fWaveHdr,SizeOf(TWaveHdr));
    // Header has now been sent
    fWaveHdr.dwUser:=1;

    SwapBuffers;

    // Unprepare oldest header
    if (fWaveHdr.dwUser <> 0) then begin
      while (waveOutUnprepareHeader(fHandle, fWaveHdr,SizeOf(TWaveHdr)) = WAVERR_STILLPLAYING) do
        sleep(SLEEPTIME);
    end;
   end else begin
    inc(FFillup);
    SwapBuffers;
    if (FFillup = 2) then begin
     // Write the previously calculated 2 headers
     waveOutPrepareHeader(fHandle,fWave1,SizeOf(TWaveHdr));
     waveOutWrite(fHandle,fWave1,SizeOf(TWaveHdr));
     fWave1.dwUser:=1;
     waveOutPrepareHeader(fHandle,fWave2,SizeOf(TWaveHdr));
     waveOutWrite(fHandle,fWave2,SizeOf(TWaveHdr));
     fWave2.dwUser:=1;
    end;
   end;
  end;
end;

procedure TWaveOut.ClearBuffer;
var i: Cardinal;
begin
{$ifdef log}WriteLn('ClearBuffer');{$endif}
  waveOutReset(fHandle);

  for i := 0 to 2 do begin
   if fHeaders[i].dwUser<>0 then begin
    waveOutUnprepareHeader(fHandle,@fHeaders[i],SizeOf(TWaveHdr));
    fHeaders[i].dwUser:=0;
   end;
   FillChar(fHeaders[i].lpData^,fDataSize,0);
  end;

  // Force the buffers to fillup before playing.
  fBuffer:=0;
  fFillup:=0;
  fBufferCount:=0;
  // Reset buffer pointers
  SwapBuffers;
end;

procedure TWaveOut.SetStopFlag;
begin
 fUserStop := 1;
end;

initialization
{$ifdef log}AllocConsole;{$endif}
end.
